from django.conf import settings
from django.test import TestCase

class TestSettings(TestCase):

    def test_inexistent_returns_default(self):
        """Assert a inexistent key returns the provided default value."""
        assert getattr(settings, 'INEXISTENT', 'something') == 'something'

    def test_only_in_core_config(self):
        """Asserts values defined only in core config."""
        assert settings.DEBUG == False

    def test_only_in_package(self):
        """Asserts default settings defined in the package."""
        assert settings.MYPACKAGEVAR == "ok"

    def test_only_in_user_config(self):
        """Asserts LDP packages are loaded from YAML file."""
        assert 'djangoldp.tests' in settings.DJANGOLDP_PACKAGES

    def test_overrided_core_by_package_config(self):
        assert settings.USE_I18N == False

    def test_overrided_package_by_user_config(self):
        assert settings.USE_TZ == False

    def test_overrided_core_by_user_config(self):
        """Asserts values overrided from user configuration."""
        assert settings.EMAIL_HOST == 'somewhere'

    def test_installed_apps_resolution(self):
        """Asserts LDP packages are referenced along with default installed apps."""
        # test inclusion from server YAML settings
        assert 'djangoldp.tests' in settings.INSTALLED_APPS
        # test inclusion from ldppackage settings
        assert 'djangoldp.tests.dummy.apps.DummyConfig' in settings.INSTALLED_APPS
        # test inclusion from default core settings
        assert 'djangoldp' in settings.INSTALLED_APPS
        # test deduplication of dummy app
        assert settings.INSTALLED_APPS.count('djangoldp.tests.dummy.apps.DummyConfig') == 1

        # FIXME: We should check the order

    def test_reference_middleware(self):
        """Asserts middlewares added in packages are added to the settings."""
        assert 'djangoldp.tests.dummy.middleware.DummyMiddleware' in settings.MIDDLEWARE

    def test_extra_module(self):
        #FIXME
        pass



# Generated by Django 2.2.17 on 2021-01-25 18:47

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djangoldp', '0014_auto_20200909_2206'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='activity',
            options={'default_permissions': ['add', 'change', 'delete', 'view', 'control']},
        ),
        migrations.AlterModelOptions(
            name='follower',
            options={'default_permissions': ['add', 'change', 'delete', 'view', 'control']},
        ),
        migrations.AlterModelOptions(
            name='ldpsource',
            options={'default_permissions': ['add', 'change', 'delete', 'view', 'control'], 'ordering': ('federation',)},
        ),
        migrations.AlterModelOptions(
            name='scheduledactivity',
            options={'default_permissions': ['add', 'change', 'delete', 'view', 'control']},
        ),
    ]



from copy import deepcopy
import random


'''
Contains definitions used in common by multiple scripts within this directory
'''


def generate_user(i, user_template):
    user = deepcopy(user_template)
    user['pk'] = i
    user['fields']['username'] = str('fixture-' + str(i))
    user['fields']['email'] = user['fields']['username'] + "@c.coop"
    return user


def generate_users(count, user_template, fixture=None, offset=0):
    if fixture is None:
        fixture = list()

    for i in range(count):
        j = offset + i
        user = generate_user(j, user_template)
        fixture.append(user)

    return fixture


def generate_project_members(project_pk, fixture, offset, total_users):
    max_members_per_project = 10

    def generate_project_member(i, user_pk):
        return {
            'model': 'djangoldp_project.member',
            'pk': i,
            'fields': {
                'project': project_pk,
                'user': user_pk
            }
        }

    # we want to select a handful of random users
    # to save time we just select a random user within a safe range and then grab a bunch of adjacent users
    start_user_pk = random.randint(max(offset, 1), offset + (total_users - (max_members_per_project + 1)))
    if start_user_pk < offset:
        return

    for i in range(random.randint(1, max_members_per_project)):
        j = offset + (i + (project_pk * max_members_per_project))  # generate a unique integer id
        user_pk = start_user_pk + i  # select the next user

        fixture.append(generate_project_member(j, user_pk))

    return fixture


def generate_project(i, project_template):
    project = deepcopy(project_template)
    project['pk'] = i

    return project


def generate_projects(count, project_template, fixture=None, offset=0):
    if fixture is None:
        fixture = list()

    for i in range(count):
        j = offset + i
        project = generate_project(j, project_template)
        fixture.append(project)

        # append random number of project members, max 10 for a single project
        generate_project_members(j, fixture, offset, count)

    return fixture



import os
import sys
import yaml
import logging
from django.core.exceptions import ImproperlyConfigured
from django.conf import settings as django_settings
from pathlib import Path
from collections import OrderedDict
from typing import Iterable
from . import default_settings

try:
    from importlib import import_module
except ImportError:
    from django.utils.importlib import import_module


logger = logging.getLogger(__name__)


def configure(filename='settings.yml'):
    """Helper function to configure django from LDPSettings."""

    yaml_config = None
    try:
        with open(filename, 'r') as f:
            yaml_config = yaml.safe_load(f)
    except FileNotFoundError:
        logger.info('Starting project without configuration file')

    # ref: https://docs.djangoproject.com/fr/2.2/topics/settings/#custom-default-settings
    ldpsettings = LDPSettings(yaml_config)

    django_settings.configure(ldpsettings)


class LDPSettings(object):

    """Class managing the DjangoLDP configuration."""

    def __init__(self, config):

        """Build a Django Setting object from a dict."""

        if django_settings.configured:
            raise ImproperlyConfigured('Settings have been configured already')

        self._config = config
        self._settings = self.build_settings()

    def build_settings(self, extend=['INSTALLED_APPS', 'MIDDLEWARE']):
        """
        Look for the parameters in multiple places.
        Each step overrides the value of the previous key found. Except for "extend" list. Those value must be lists and all values found are added to these lists without managing duplications.

        Resolution order of the configuration:
          1. Core default settings
          2. Packages settings
          3. Code from a local settings.py file
          4. YAML config file
        """

        # helper loop
        def update_with(config):
            for k, v in config.items():

                if k in extend:
                    settings[k].extend(v)

                elif not k.startswith('_'):
                    settings.update({k: v})

        # start from default core settings
        settings = default_settings.__dict__.copy()
        logger.debug(f'Building settings from core defaults')

        # INSTALLED_APPS starts empty
        settings['INSTALLED_APPS'] = []

        # look settings from packages in the order they are given (local overrides installed)
        for pkg in self.DJANGOLDP_PACKAGES:

            # FIXME: There is something better to do here with the sys.modules path
            try:
                # override with values from installed package
                mod = import_module(f'{pkg}.djangoldp_settings')
                update_with(mod.__dict__)
                logger.debug(f'Updating settings from installed package {pkg}')
            except ModuleNotFoundError:
                pass

            try:
                # override with values from local package
                mod = import_module(f'{pkg}.{pkg}.djangoldp_settings')
                update_with(mod.__dict__)
                logger.debug(f'Updating settings from local package {pkg}')
            except ModuleNotFoundError:
                pass

        # look in settings.py file in directory
        try:
            mod = import_module('settings')
            update_with(mod.__dict__)
            logger.debug(f'Updating settings from local settings.py file')
        except ModuleNotFoundError:
            pass

        # look in YAML config file 'server' section
        try:
            conf = self._config.get('server', {})
            update_with(conf)
            logger.debug(f'Updating settings with project config')
        except KeyError:
            pass

        # In the end adds the INSTALLED_APPS from the core
        settings['INSTALLED_APPS'].extend(getattr(default_settings,'INSTALLED_APPS'))

        return settings

    @property
    def DJANGOLDP_PACKAGES(self):

        """Returns the list of LDP packages configured."""

        pkg = self._config.get('ldppackages', [])
        return [] if pkg is None else pkg

    @property
    def INSTALLED_APPS(self):

        """Return the installed apps and the LDP packages."""

        # get ldp packages (they are django apps)
        apps = self.DJANGOLDP_PACKAGES.copy()

        # add the default apps
        apps.extend(self._settings['INSTALLED_APPS'])

        # As settings come from different origins duplicuation is likeliy to happen
        return list(OrderedDict.fromkeys(apps))

    def __getattr__(self, param):
        """Return the requested parameter from cached settings."""
        if param.startswith('_') or param.islower():
            # raise the django exception for inexistent parameter
            raise AttributeError(f'"{param}" is not compliant to django settings format')
        try:
            return self._settings[param]
        except KeyError:
            # raise the django exception for inexistent parameter
            raise AttributeError(f'no "{param}" parameter found in settings')

